import os.path
from croydon.baseapp import BaseApp


class App(BaseApp):

    def __init__(self):
        app_dir = os.path.abspath(os.path.dirname(__file__))
        project_dir = os.path.abspath(os.path.join(app_dir, ".."))
        super().__init__(project_dir=project_dir)


app = App()



from typing import Optional, Any, Protocol, TypeVar


class AbstractCache(Protocol):
    NAME: str

    async def initialise(self) -> None: ...

    async def get(self, key: str) -> Optional[Any]: ...

    async def set(self, key: str, value: Any) -> None: ...

    async def has(self, key: str) -> bool: ...

    async def delete(self, key: str) -> bool: ...



import asyncio
from abc import abstractmethod, ABC
from .types import TBaseTask
from .. import ctx


class BaseWorker(ABC):

    @abstractmethod
    async def run_task(self, task: TBaseTask) -> None: ...

    async def run(self):
        try:
            async for task in ctx.queue.tasks():
                ctx.log.debug("worker accepted task %s", task)
                await self.run_task(task)
        except asyncio.CancelledError:
            pass



from unittest import TestCase
from ..models.base_model import BaseModel
from ..models.fields import StringField
from ..models.index import Index, IndexDirection, IndexKey


class TestBaseModel(TestCase):

    def test_empty(self):
        model = BaseModel()
        self.assertIsNone(model.id)
        self.assertCountEqual(model._fields, ["id"])

    def test_not_empty(self):
        class Model(BaseModel):
            field = StringField()

        model = Model({"field": "value"})
        self.assertEqual(model.field, "value")
        self.assertCountEqual(model._fields, ["id", "field"])

    def test_collection_name(self):
        class Model(BaseModel):
            COLLECTION = "models"

        bm = BaseModel()
        self.assertEqual(bm.collection, "base_model")

        m = Model()
        self.assertEqual(m.collection, "models")

    def test_rejected_fields(self):
        class Model(BaseModel):
            string_f = StringField()
            string_rf = StringField(rejected=True)

        model = Model()
        self.assertCountEqual(model._fields, ["id", "string_f", "string_rf"])
        self.assertTrue(model._fields["string_rf"].rejected)

    def test_restricted_fields(self):
        class Model(BaseModel):
            string_f = StringField()
            string_rf = StringField(restricted=True)

        model = Model()
        self.assertCountEqual(model._fields, ["id", "string_f", "string_rf"])
        self.assertTrue(model._fields["string_rf"].restricted)

    def test_indexes(self):
        class Model(BaseModel):
            string_f = StringField()
            string_if = StringField(index=True)
            string_uf = StringField(unique=True)
            string_duf = StringField(index=IndexDirection.DESCENDING, unique=True)

        model = Model()

        self.assertCountEqual(
            model._indexes,
            [
                Index(keys=[IndexKey(key="string_if", spec=IndexDirection.ASCENDING)]),
                Index(
                    keys=[IndexKey(key="string_uf", spec=IndexDirection.ASCENDING)],
                    options={"unique": True}
                ),
                Index(
                    keys=[IndexKey(key="string_duf", spec=IndexDirection.DESCENDING)],
                    options={"unique": True}
                ),
            ],
        )

    def test_to_dict(self):
        class Model(BaseModel):
            field1 = StringField()
            field2 = StringField()
            field3 = StringField(restricted=True)

        model = Model({"field1": "value1", "field2": "value2", "field3": "value3"})

        self.assertDictEqual(
            model.to_dict(), {"id": None, "field1": "value1", "field2": "value2"}
        )

        self.assertDictEqual(
            model.to_dict(fields=["field1", "field2", "field3"]),
            {"field1": "value1", "field2": "value2"},
        )

        self.assertDictEqual(
            model.to_dict(include_restricted=True),
            {"id": None, "field1": "value1", "field2": "value2", "field3": "value3"},
        )

        self.assertDictEqual(
            model.to_dict(fields=["field1", "field3"], include_restricted=True),
            {"field1": "value1", "field3": "value3"},
        )

        self.assertDictEqual(
            model.to_dict(
                fields=["field1", "field3", "bizzare"], include_restricted=True
            ),
            {"field1": "value1", "field3": "value3"},
        )



from uuid import uuid4
from typing import Callable, Any, List
from functools import wraps
from datetime import datetime
from bson.objectid import ObjectId, InvalidId
from .errors import ObjectSaveRequired
from .types import KwArgs


def uuid4_string() -> str:
    return str(uuid4())


def save_required(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(self, *args: List[Any], **kwargs: KwArgs) -> Any:
        if self.is_new:
            raise ObjectSaveRequired("This object must be saved first")
        return func(*args, **kwargs)

    return wrapper


def resolve_id(id_: str | ObjectId | None) -> ObjectId | str | None:
    # ObjectId(None) generates a new unique object id
    # We need to handle that case and return None instead
    if id_ is not None:
        if not isinstance(id_, ObjectId):
            try:
                objid_expr = ObjectId(id_)
                if str(objid_expr) == id_:
                    return objid_expr
            except (InvalidId, TypeError):
                pass
    return id_


# Mongo stores datetime rounded to milliseconds as its datetime
# capabilities are limited by v8 engine
def now() -> datetime:
    dt = datetime.utcnow()
    dt = dt.replace(microsecond=dt.microsecond // 1000 * 1000)
    return dt


NilObjectId: ObjectId = ObjectId("000000000000000000000000")



import inspect

from typing import Any, Optional, Iterable
from .fields import Field, FieldProto
from .index import Index, IndexKey


def snake_case(name: str) -> str:
    result = ""
    for i, l in enumerate(name):
        if 65 <= ord(l) <= 90:
            if i != 0:
                result += "_"
            result += l.lower()
        else:
            result += l
    return result


TNamespace = dict[str, Any]


class DefaultCollectionMeta(type):
    @classmethod
    def __prepare__(
        mcs,
        name: str,
        bases: tuple[type[object]],
        **kwargs: dict[str, Any]
    ) -> TNamespace:
        return {
            "COLLECTION": snake_case(name)
        }


class MetaModel(metaclass=DefaultCollectionMeta):

    _fields: dict[str, FieldProto]
    _indexes: Iterable[Index]
    _cache_key_fields: set[str]

    collection: str

    # You will usually need INDEXES only for defining compound ones, as
    # simple single-field indexes including "unique" have a shortcut arg
    # in a field definition, i.e.
    #
    # username = StringField(unique=True)
    # owner_id = ReferenceField(index=True)
    INDEXES: Iterable[Index] | None = None

    COLLECTION: str = ""
    KEY_FIELD: str = "_id"
    CACHE_KEY_FIELDS: Optional[Iterable[str]] = None

    def __init_subclass__(cls) -> None:
        cls._fields = {
            name: obj
            for name, obj in inspect.getmembers(cls)
            if isinstance(obj, Field)
        }
        cls._indexes = cls.__get_indexes()
        cls._cache_key_fields = cls.__get_cache_key_fields()
        cls.collection = cls.COLLECTION  # TODO: Remove one of the two

    @classmethod
    def __get_indexes(cls) -> Iterable[Index]:
        for field, descriptor in cls._fields.items():
            if descriptor.index is None:
                continue
            yield Index(
                keys=[IndexKey(key=field, spec=descriptor.index)],
                options=descriptor.index_options,
            )

        seen = set()
        for base in reversed(inspect.getmro(cls)):
            if not issubclass(base, MetaModel):
                continue
            if id(base.INDEXES) not in seen and base.INDEXES:
                seen.add(id(base.INDEXES))
                yield from base.INDEXES

    @classmethod
    def __get_cache_key_fields(cls) -> set[str]:
        cache_fields = set()
        for base in inspect.getmro(cls):
            if not issubclass(base, MetaModel):
                continue
            cache_fields.add(base.KEY_FIELD)
            if base.CACHE_KEY_FIELDS:
                cache_fields.update(base.CACHE_KEY_FIELDS)
        return cache_fields



import tomllib
from typing import Literal, Optional, Self, List
from pydantic import BaseModel, Field
from datetime import timedelta
from .types import CacheType, LogLevel


_DEFAULT_TIMEOUT = timedelta(seconds=1, milliseconds=100)
_DEFAULT_COOKIE_TTL = timedelta(days=90)


class SessionConfig(BaseModel):
    cookie: str = "_croydon_session_id"
    ttl: timedelta = _DEFAULT_COOKIE_TTL


class DatabaseConfig(BaseModel):
    host: str = "127.0.0.1"
    port: int = Field(3000, gt=0, lt=65536)
    dbname: str = "croydondb"
    timeout: timedelta = _DEFAULT_TIMEOUT

    @property
    def uri(self) -> str:
        return f"mongodb://{self.host}:{self.port}/{self.dbname}"


class LogConfig(BaseModel):
    level: LogLevel = "debug"
    filename: Optional[str] = None
    stdout: bool = True


class QueueConfig(BaseModel):
    type: Literal["mongo"] = "mongo"


class MongoQueueConfig(BaseModel):
    tasks_collection: str = "croydon_tasks"
    keep_done_tasks_for: timedelta = timedelta(days=30)
    cleanup_interval: timedelta = timedelta(hours=1)


class CacheConfig(BaseModel):
    level1: CacheType = "simple"
    level2: CacheType = "memcached"


class MemcachedConfig(BaseModel):
    backends: List[str] = ["127.0.0.1:11211"]


class OAuthConfig(BaseModel):
    id: str
    secret: str
    authorize_url: str


class WebConfig(BaseModel):
    host: str = "127.0.0.1"
    port: int = Field(3000, gt=0, lt=65536)


class GeneralConfig(BaseModel):
    documents_per_page: int = 20
    enable_passwords: bool = True


class BaseConfig(BaseModel):
    session: SessionConfig = Field(default_factory=SessionConfig)
    database: DatabaseConfig = Field(default_factory=DatabaseConfig)
    logging: LogConfig = Field(default_factory=LogConfig)
    oauth: Optional[OAuthConfig] = Field(None)
    web: WebConfig = Field(default_factory=WebConfig)
    general: GeneralConfig = Field(default_factory=GeneralConfig)
    queue: QueueConfig = Field(default_factory=QueueConfig)
    mongo_queue: MongoQueueConfig = Field(default_factory=MongoQueueConfig)
    cache: CacheConfig = Field(default_factory=CacheConfig)
    memcached: MemcachedConfig = Field(default_factory=MemcachedConfig)

    @classmethod
    def parse(cls, filename) -> Self:
        with open(filename, "rb") as f:
            config = tomllib.load(f)
        return cls(**config)
